from asyncore import read
from utility.privilege_manager import *

manager                   = Privilege_Manager()
inventory_read_only_role  = Privilege_Role(name="read_only")
inventory_admin_role      = Privilege_Role(name="admin")


def init() :

    manager.create_domain(name="inventory",description="privilege domain for inventory")

    inventory_privilege_domain = manager.get_domain_by_name("inventory")
    inventory_privilege_domain.add_role(inventory_read_only_role)
    inventory_privilege_domain.add_role(inventory_admin_role)

    inventory_privilege_domain.include_role(inventory_admin_role,inventory_read_only_role)




def populate_manager() :
    None
